"""
$RCSfile: XMLTransform.py,v $

XMLTransform automatically transforms an XML document via XSLT, where
the XML document is obtained from another Zope object (the "source"
object) via acquisition.

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/11/01 05:10:30 $'[6:-2]
__version__ = '$Revision: 1.45 $'[10:-2]

# Base classes, peer classes
from OFS.SimpleItem import SimpleItem  # The SimpleItem module is a Zope builtin that provides
                                       # cut/paste, management views, webdav/ftp, undo, ownership,
                                       # traversal, acquisition, and persistence
from Products.ZCatalog.CatalogAwareness import CatalogAware # Reqd for ZCatalog support
from OFS.PropertyManager import PropertyManager  # Reqd for proper property mgmt

# peer classes
from IXMLTransform import IXMLTransform

# Zope builtins
from Globals import DTMLFile, MessageDialog
from Acquisition import aq_chain, aq_get
from AccessControl import ClassSecurityInfo
import OFS      # we grab his _property settings
import Globals  # InitializeClass (security stuff)

# Python builtins
import mimetypes

################################################################
# Defaults
################################################################

try:
    from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
    behave_like_list = [ OFS.DTMLMethod.DTMLMethod.meta_type,
                         ZopePageTemplate.meta_type ]
except:
    behave_like_list = [ OFS.DTMLMethod.DTMLMethod.meta_type ]
    
################################################################
# Module Scoped Convenience Methods
################################################################

def addInstance(folder,
                id, title='', description='',                 # standard
                transformerName='', sourceID='', cache='on'): # xmltransform
    """
    
    This is a convenience factory method for creating an instance of
    XMLTransform.  It returns the object created, and may therefore be
    more convenient than the addXMLTransform() method it calls.  It is
    used by the unit testing programs.

    """
    folder.manage_addProduct['XMLTransform'].addXMLTransform(id,title,description,
                                                             transformerName,sourceID)
    return folder[id]

################################################################
# Container Methods
################################################################

manage_addXMLTransformForm = DTMLFile('dtml/create', globals())
xsltregistry_formpart      = DTMLFile('dtml/xsltregistry_formpart', globals())
cachemanager_formpart      = DTMLFile('dtml/cachemanager_formpart', globals())
xmltransform_formpart      = DTMLFile('dtml/xmltransform_formpart', globals())
create_standard_formpart   = DTMLFile('dtml/create_standard_formpart', globals())

def findTransformerRegistry(self):
    """

    Find an instance of TransformerRegistry to obtain transformer
    information.
    
    All XMLTransform instances require an instance of TransformerRegistry.
    The current policy is as follows: use the first registry found
    by acquisition.  Alternatively, a more complex policy could be
    used such as 
    a) looking in some predefined place, where the place could be 
    specified in some property (a la Zope3 explicit acquisition), 
    or 
    b) finding all registries in the ZODB and presenting a list for 
    selection by the user.
    For right now, we don't want to over-engineer the damn thing. 
    CKS 7/24/2002.
      
    """

    # Note: superValues does _NOT_ find things in the "Root Folder"
    metaType   = 'XSL Transformer Registry'
    registries = self.superValues(metaType)
    if registries:
        return registries[0]
    else:
        return None

def findCacheManager(self):
    """

    Find an instance of CacheManager to use (optionally, provided
    the caching property is set to 'on')

    CacheManager is purely optional.  If it is present, caching may be
    done.  If applicable, the XMLTransform will use the CacheManager
    it finds via this method
        
    The current policy is as follows: use the first cache manager
    found by acquisition.  Alternatively, a more complex policy could
    be used such as
    a) looking in some predefined place, where the place could be
    specified in some property (a la Zope3 explicit acquisition), or
    b) finding all cache managers in the ZODB and presenting a list
    for selection by the user.
    For right now, we don't want to over-engineer the damn thing.  CKS
    8/4/2002.
      
    """

    # Note: superValues does _NOT_ find things in the "Root Folder"
    metaType      = 'XML Transform Cache Manager'
    cacheManagers = self.superValues(metaType)
    if cacheManagers:
        return cacheManagers[0]
    else:
        return None

def manage_addXMLTransform(self,
                           id, title='', description='',          # standard
                           transformerName='',sourceID='',cache='on',# xmltransform
                           REQUEST=None, RESPONSE=None):
    """

    Factory method to create an instance of XMLTransform, called from
    a GUI in the Zope Management Interface.  It calls addXMLTransform
    to actually do the work.

    """
    try:
        self.addXMLTransform(id, title, description,
                             transformerName, sourceID,
                             cache)
        message = 'successfully created ' + id + ' XMLTransform object.'        
        return MessageDialog(title   = 'CREATED',
                             message = message,
                             action  = 'manage_main')
        
    except Exception, e:
        message = str(e)
        message.replace('\n','<br/>')
        return MessageDialog(title   = 'ERROR',
                             message = message,
                             action  = 'manage_main')

def addXMLTransform(self,
                    id, title='', description='',                 # standard
                    transformerName='', sourceID='', cache='on'): # xmltransform
    """

    Factory method to actually create an instance of XMLTransform and
    return it.

    You should call this method directly if you are creating an
    instance of XMLTransform programatically.

    """

    if not id or not sourceID or not transformerName:
        raise Exception('Required fields must not be blank')

    # does the transformerName point to a valid transformer?
    reg   = self.findTransformerRegistry()
    if reg is None:
        raise Exception('Transformer Registry not found')
    tran = reg.transformerObject(transformerName)
    if tran is None:
        message = "Invalid transformerName %s" % (transformerName)
        raise Exception(message)
    
    # if transformer is a ZPT, we should also behave_like a ZPT
    behave_like = None
    try:
        from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
        if tran.meta_type == ZopePageTemplate.meta_type:
            behave_like = tran.meta_type
    except:
        pass
    
    # does the sourceID point to a valid Zope object?
    if aq_get(self, sourceID, None) is None:
        message = "I cannot find the %s object.\nIt is not in my acquisition path." \
                  % (sourceID)
        raise Exception(message)

    self._setObject(id, XMLTransform(id, title, description,
                                     transformerName, sourceID,
                                     behave_like, cache))
    self._getOb(id).reindex_object()

################################################################
# XMLTransform class
################################################################

class XMLTransform(CatalogAware,    # ZCatalog support
                   PropertyManager, # Property support
                   SimpleItem):
    """

    Automatically transforms an XML document via XSLT, where the XML
    document is obtained from another Zope object via acquisition.

    """
    meta_type = 'XML Transform'

    __implements__ = IXMLTransform

    # This tuple defines our properties
    _properties = (
        {'id':'title',          'type':'string',    'mode': 'w'},
        {'id':'description',    'type':'string',    'mode': 'w'},
        {'id':'debugLevel',     'type':'int',       'mode': 'w'},
        {'id':'caching',        'type':'selection', 'mode': 'w', \
         'select_variable':"onOff"},
        {'id':'content_type',   'type':'selection', 'mode': 'w', \
         'select_variable':'getMimeTypes', },

        {'id':'behave_like',    'type':'selection', 'mode': 'w', \
         'select_variable':'getBehaveListList', },        
        )
    
    manage_options = (
        {'label':'Edit',    'action':'manage_editForm',    'help': ('XMLTransform','edit.stx') },
        {'label':'View',    'action':''}, ) + \
        OFS.PropertyManager.PropertyManager.manage_options + \
        OFS.SimpleItem.SimpleItem.manage_options

    _security = ClassSecurityInfo()

    # set security for the object itself, e.g. if it is accessed in DTML code
    # This line is REQUIRED to allow access to XMLTransforms within public-viewable DTML code
    _security.declareObjectPublic()

    # call __call__() method on DTMLFile for our edit form, point at our edit_form.dtml
    # globals() is standard python for passing the current namespace
    _security.declareProtected('Manage XML Transforms',      'manage_editForm')
    manage_editForm           = DTMLFile('dtml/edit_form',globals())

    # this exists so we can call it from DTML
    xsltregistry_formpart  = xsltregistry_formpart
    cachemanager_formpart  = cachemanager_formpart    
    # end of DTML scoping hack

    # This is used for verification messages like "File Changed"
    # when you hit the button in the edit form.
    manage_header_message = DTMLFile('dtml/manage_header_message', globals())

    def __init__(self, id, title, description,
                 transformerName, sourceID,
                 behave_like = None,
                 caching='on', debugLevel=0):
        """

        Initialize a new XMLTransform.  All of the arguments after
        'self' are strings.

        """

        self.id              = id
        self.title           = title
        self.description     = description
        self.__version__     = __version__
        self.transformerName = transformerName        
        self.sourceID        = sourceID
        if behave_like:
            self.behave_like = behave_like
        else:
            self.behave_like = behave_like_list[0]
        self.caching         = caching
        self.isCacheDirty    = 1
        self.debugLevel      = debugLevel

        #
        # By default, produce HTML
        #
        self.content_type    = 'text/html'

    ################################################################
    # Methods called from ZMI pages
    ################################################################

    _security.declareProtected('Edit',     'manage_editSource')
    def manage_editSource(self, REQUEST, RESPONSE):
        "Call manage_edit on underlying XML source object"

        source = self.findSourceObject()
        RESPONSE.redirect( source.absolute_url() + "/" + "manage_workspace" )

    _security.declareProtected('Edit',     'manage_editTransform')
    def manage_editTransform(self, transformerName, sourceID,
                             REQUEST, RESPONSE):
        "Change transformer to be used or ID of source object"

        # does the source ID point to a valid Zope object?
        try:
            self.editTransform(transformerName, sourceID)
        except Exception, e:
            return MessageDialog(title   = 'ERROR',
                                 message = str(e),
                                 action  = 'manage_editForm')

        message = 'XMLTransform edits saved.'
        return self.manage_editForm(header_message=message)

    _security.declareProtected('Edit', 'manage_editProperties')
    def manage_editProperties(self, REQUEST):
        """

        Cover for PropertyManager.manage_editProperties() method.  set
        debugLevel of underlying XSLTProcessor to our debugLevel, then
        pass it on to the inherited method for further processing

        """
        for prop in self._propertyMap():
            name = prop['id']
            if name == 'debugLevel':
                value = REQUEST.get(name, '')
                self.setDebugLevel(value)
        return XMLTransform.inheritedAttribute("manage_editProperties")(self, REQUEST)

    ################################################################
    # Methods implementing the IXMLTransform interface below
    ################################################################

    _security.declareProtected('View',  'isCacheFileUpToDate')
    def isCacheFileUpToDate(self, REQUEST=None):
        """

        Return 1 if and only if self should use the cached value
        rather than regenerating the transformed value, and the cache
        manager exists, and the cache file exists.  Note: this
        algorithm is rather simple and limited right now, but we
        intend to improve it over time.  Many items are not taken into
        account today: a) XML fragments that are included into the
        main document via the XSLT document() function b) XSLT
        transformer parameters c) XSLT transformers themselves (as
        well as all included and imported transformers).  FIXME cks
        11/26/2001

        """
        # isCacheDirty will be true if we have switched xforms or source docs
        if self.isCacheDirty:
            return 0

        manager = self.findCacheManager()
        if manager is None:
            return 0
        
        srcObject = self.findSourceObject()
        
        reg = self.findRegistry()
        if reg is None:
            raise Exception('Transformer Registry not found')
        xformObject = reg.transformerObject(self.transformerName)
        
        # ZODB mod time.  This accounts for properties of the
        # XMLTransform Zope object itself, and its content, unless
        # its content is stored in an external file.
        # NOTE: this value is returned as number of seconds
        # since the epoch in UTC (see python time module)
        #
        # Unfortunately, the Python time module returns number of
        # seconds, truncated.  bobobase returns microseconds.
        # Therefore, we must truncate by subtracting 0.5 and rounding.
        sourceTime = round(srcObject.bobobase_modification_time().timeTime() - 0.5)
        xformTime  = round(xformObject.bobobase_modification_time().timeTime() - 0.5)        
        cacheTime  = manager.cacheFileTimeStamp(self)

        return ( (cacheTime >= sourceTime) and (cacheTime >= xformTime) )

    _security.declareProtected('View',  'findRegistry')
    def findRegistry(self):
        """

        Find an instance of TransformerRegistry to obtain sandbox
        information.  This calls a module scoped function which does
        the real work of finding the sandbox.  Because we need this
        behavior to even _create_ an instance of XMLTransform the
        function needs to be module scoped.

        """
        return findTransformerRegistry(self)

    _security.declareProtected('View',  'findCacheManager')
    def findCacheManager(self):
        """

        Find an instance of CacheManager to use (optionally, provided
        the caching property is set to 'on') This calls a module
        scoped function which does the real work of finding the
        sandbox.  Because we need this behavior to even _create_ an
        instance of XMLTransform the function needs to be module
        scoped.
      
        """
        return findCacheManager(self)
    
    _security.declareProtected('View', 'findSourceObject')
    def findSourceObject(self):
        "Retrieve the source object by using acquisition on the ID"

        if self.isDebugging():
            print "Requesting contents of ", self.sourceID
        return aq_get(self.getSelf(), self.sourceID, None)

    _security.declareProtected('View', 'getContents')
    def getContents(self, REQUEST):
        """
        
        Return the source of the object this XMLTransform is pointing
        to as a string
        
        """
        source_object = self.findSourceObject()
        return source_object(self.getSelf(), REQUEST)

    _security.declareProtected('View', 'transform')
    def transform(self, REQUEST):
        """

        Generate result using transformer and return it as a string

        """

        registry    = self.findRegistry()
        processor   = registry.processor()
        xslObject   = registry.transformerObject(self.transformerName)
        xslContents = xslObject(xslObject, REQUEST)
        xslURL      = xslObject.absolute_url()
        xmlObject   = self.findSourceObject()
        xmlContents = xmlObject(self.getSelf(), REQUEST)
        xmlURL      = xmlObject.absolute_url()

        processor.setDebugLevel(self.debugLevel)
        
        return processor.transform(xmlContents,
                                   xmlURL,
                                   xslContents,
                                   xslURL,
                                   self, REQUEST)

    _security.declareProtected('View', 'testTransform')
    def testTransform(self, REQUEST):
        """

        This version is used purely for testing.  It is the same as
        transform() but it throws exceptions.  This is used for
        negative tests.  See tests/TestXMLTransform()

        """

        registry       = self.findRegistry()
        processor      = registry.processor()
        xslObject      = registry.transformerObject(self.transformerName)
        xslContents    = xslObject(xslObject, REQUEST)
        xslURL         = xslObject.absolute_url()
        xmlObject      = self.findSourceObject()
        xmlContents    = xmlObject(self.getSelf(), REQUEST)
        xmlURL         = xmlObject.absolute_url()
        topLevelParams = processor.getXSLParameters(self)

        processor.setDebugLevel(self.debugLevel) 

        return processor.transformGuts(xmlContents,
                                       xmlURL,
                                       xslContents,
                                       xslURL,
                                       self,
                                       topLevelParams,
                                       REQUEST)

    _security.declareProtected('View',     'transformAndCache')
    def transformAndCache(self, REQUEST):
        """

        Generate results using transformer and save them to the
        cache, if applicable

        """

        rawResult = self.transform(REQUEST)
        if self.isCachingOn():
            manager = self.findCacheManager()
            if manager is not None:
                manager.saveToCache(self, rawResult)
                self.isCacheDirty = 0

    _security.declareProtected('Edit',     'setCachingOn')
    def setCachingOn(self, REQUEST=None):
        "same as changing the property.  For use in scripts or by the cache Manager"
        self.caching = 'on'

    _security.declareProtected('Edit',     'setCachingOff')
    def setCachingOff(self, REQUEST=None):
        "same as changing the property.  For use in scripts or by the cache Manager"
        self.caching = 'off'

    # innocuous methods should be declared public
    _security.declarePublic('isCachingOn')
    def isCachingOn(self):
        "Return TRUE if caching is turned on"
        return self.caching == 'on'

    # innocuous methods should be declared public
    _security.declarePublic('isDebugging')
    def isDebugging(self):
        "Return true if and only if debugging is on."
        return self.debugLevel > 0

    _security.declareProtected('Edit',     'manage_editTransform')
    def editTransform(self, transformerName, sourceID):
        "Change transformer to be used or ID of source object"

        # does the source ID point to a valid Zope object?
        if aq_get(self, sourceID, None) is None:
            raise Exception("""I cannot find the %s object.\n
            It is not in my acquisition path.""" % (sourceID))

        # does the transformerName point to a valid transformer?
        reg   = self.findRegistry()
        if reg is None:
            raise Exception('Transformer Registry not found')
        tran = reg.transformerObject(transformerName)
        if tran is None:
            message = "Invalid transformerName %s" % (transformerName)
            raise Exception(message)
        
        self.sourceID        = sourceID
        self.transformerName = transformerName
        self._p_changed      = 1  # dirty flag, triggers persistence to update the object
        self.isCacheDirty    = 1

    ################################################################
    # Utilities
    ################################################################

    _security.declarePublic('getSelf')
    def getSelf(self):
        "Return this object. For use in DTML scripts"
        return self.aq_chain[0]

    _security.declarePublic('onOff')
    def onOff(self):
        return ["on", "off"]

    _security.declarePublic('getBehaveLikeList')
    def getBehaveListList(self):
        "return list of standard zope objects this XMLTransform can behave like"
        return behave_like_list

    _security.declarePublic('getMimeTypes')
    def getMimeTypes(self):
        "return list of available mime types."
        typelist = mimetypes.types_map.values()

        # return sorted, unique-ified list.
        # Code stolen from python FAQ answer 4.64
        typelist.sort()
        last = typelist[-1]
        for i in range(len(typelist)-2, -1, -1):
            if last==typelist[i]: del typelist[i]
            else: last=typelist[i]
        return typelist

    _security.declarePublic('setDebugLevel')
    def setDebugLevel(self, value):
        "Set debug level for ourselves and our underlying processor"
        self.debugLevel = value

    ################################################################
    # Standard Zope stuff
    ################################################################

    _security.declareProtected('View', '__str__')
    def __str__(self, REQUEST = None):
        "return contents as string"
        message = self.id + " (" + self.title + ")"
        return message

    # We inherit __len__ from one of our base classes
    # BTW, __len__ should NEVER return 0, or Zope will assume the
    # object does "not exist" and lines like "if myObject:" will fail.

    # No PUT editing via FTP or webdav is disallowed.
    
    _security.declareProtected('FTP access', 'manage_FTPget')
    manage_FTPget = __str__

    _security.declareProtected('Access contents information', 'index_html')
    # next line is not strictly necessary, Access contents info is Anonymous by default
    # as opposed to all other protections that get Manager Role by default
    _security.setPermissionDefault('Access contents information', ('Anonymous')) 
    def index_html(self, REQUEST = None, RESPONSE = None):
        """Default view of rendered version of XML Document
        The View tab in a standard management view calls index_html."""

        return "<html><body>" + self(self, REQUEST) + "</body></html>"
        
    _security.declareProtected('View', '__call__')
    def __call__(self, client = None, REQUEST = None, RESPONSE = None):
        "Render self by processing its content with the named XSLT stylesheet"

        # Check for caching
        rawResult = None
        if self.isCachingOn():
            if self.isCacheFileUpToDate():
                manager = self.findCacheManager()
                if manager is not None:
                    rawResult = manager.valueFromCache(self)

        if rawResult is None:
            rawResult = self.transform(REQUEST)
            manager = self.findCacheManager()            
            if self.isCachingOn() and manager is not None:
                manager.saveToCache(self, rawResult)
                self.isCacheDirty = 0

	# render as DTML or page template

        if self.behave_like == OFS.DTMLMethod.DTMLMethod.meta_type:
            resultDocument = OFS.DTMLMethod.DTMLMethod(rawResult)
            resultDocument.id           = self.id
            resultDocument.title        = self.title            
            resultDocument.content_type = self.content_type

            # Call resultDocument within the context of the client so that
            # acquisition works as expected.  That displays the
            # resultDocument and causes any embedded DTML tags to be
            # resolved as well.  The embedded DTML tags could refer to other
            # XMLTransforms which themselves will be rendered, and so on,
            # recursively.
            if client:
                return resultDocument.__of__(client)(client, REQUEST)
            else:
                return resultDocument.__of__(self)(self, REQUEST)
            
        else: # PageTemplate
            try:
                from Products.PageTemplates.ZopePageTemplate import ZopePageTemplate
                resultDocument = ZopePageTemplate(
                    self.id,
                    rawResult,
                    self.content_type)
                resultDocument.pt_setTitle(self.title)

                # Call resultDocument within the context of the client so that
                # acquisition works as expected.  That displays the
                # resultDocument and causes any embedded page templates to be
                # resolved as well.  The embedded templates could refer to other
                # XMLTransforms which themselves will be rendered, and so on,
                # recursively.
                if client:
                    return resultDocument.__of__(client)(client, REQUEST)
                else:
                    return resultDocument.__of__(self)(self, REQUEST)
            
            except:
                # no support for page templates.  This should never happen...
                pass

    ################################################################
    # Support for Schema Migration
    ################################################################

    def repair(self):
        """

        Repair this object.  This method is used for schema migration,
        when the class definition changes and existing instances of
        previous versions must be updated.

        """
        if not hasattr(self,'isCacheDirty'):
            self.isCacheDirty = 1
        self.behave_like = behave_like_list[0]

        if hasattr(self, 'processor'):
            self.processor = None

# register security information
Globals.InitializeClass(XMLTransform)
