"""
$RCSfile: TestTransformerRegistry.py,v $
Unit tests for XMLTransform product

Author: <a href="mailto:cstrong@arielpartners.com">Craeg Strong</a>
Release: 1.0
"""

__cvstag__  = '$Name:  $'[6:-2]
__date__    = '$Date: 2002/10/30 21:40:13 $'[6:-2]
__version__ = '$Revision: 1.7 $'[10:-2]

import os, sys
if __name__ == '__main__':
    execfile(os.path.join(sys.path[0], 'framework.py'))

# Load fixture
from Testing import ZopeTestCase

# XMLTransform
from Products.XMLTransform import TransformerRegistry

ZopeTestCase.installProduct('XMLTransform')

class TransformerRegistryTestCase(ZopeTestCase.ZopeTestCase):
    """

    Transformer Registry unit tests for the XMLTransform Zope product.
    These tests use the ZopeTestCase enhanced testing product.

    """

    ################################################################
    # Fixture
    ################################################################

    def afterSetUp(self):
        '''Add object to default fixture'''
        self.reg = TransformerRegistry.addInstance(self.folder, 'registry')

    def afterClear(self):
        '''Clean up after myself'''
        try: del self.reg
        except AttributeError: pass

    ################################################################
    # Test Cases
    ################################################################

    def test_01(self):
        "Ensure that Registry cannot be created with a blank ID"
        self.assertRaises(Exception,
                          TransformerRegistry.addInstance,
                          self.folder, '')

    def test_02(self):
        "Ensure that test fixture is properly setup"
        self.assertEquals(self.reg.id, 'registry')
        self.assertEquals(self.reg.meta_type,
                          TransformerRegistry.TransformerRegistry.meta_type)
        
    def test_03(self):
        "Ensure that an empty registry returns no transformers"
        names = self.reg.transformerNames()
        self.assertEquals(len(names),0)

    def test_04(self):
        "Ensure that a non empty registry returns the names of its transformers"
        names = self.reg.transformerNames()
        self.assertEquals(len(names),0)
        self.reg.manage_addDTMLMethod('aTransformer')
        names = self.reg.transformerNames()
        self.assertEquals(len(names),1)
        self.assertEquals(names[0],'aTransformer')

        self.reg.manage_addFolder('aFolder')
        folder = self.reg['aFolder']
        folder.manage_addDTMLMethod('secondTransformer')
        names = self.reg.transformerNames()
        self.assertEquals(len(names),2)
        self.failUnless('aTransformer' in names)
        self.failUnless('aFolder/secondTransformer' in names)

    def test_05(self):
        "Ensure that objects that are not of a transformer_meta_type are omitted"
        names = self.reg.transformerNames()        
        self.assertEquals(len(names),0)
        self.reg.manage_addImage('myimage', 'testfiles/XMLTransform.gif')
        names = self.reg.transformerNames()        
        self.assertEquals(len(names),0)

    def test_06(self):
        "Ensure transformers can be retrieved by name"
        names = self.reg.transformerNames()        
        self.assertEquals(len(names),0)
        self.reg.manage_addDTMLMethod('aTransformer')
        xform = self.reg['aTransformer']
        names = self.reg.transformerNames()
        self.assertEquals(len(names),1)
        self.assertEquals(names[0],'aTransformer')

        xform2 = self.reg.transformerObject('aTransformer')
        self.assertEquals(xform,xform2)

        xform3 = self.reg.transformerObject('idontexist')
        self.assertEquals(xform3, None)

    def test_07(self):
        "Test a highly nested transformer"
        names = self.reg.transformerNames()        
        self.assertEquals(len(names),0)

        self.reg.manage_addFolder('A')
        folderA = self.reg['A']
        folderA.manage_addFolder('B')
        folderB = folderA['B']        
        folderB.manage_addFolder('C')
        folderC = folderB['C']        
        folderC.manage_addFolder('D')
        folderD = folderC['D']        
        folderD.manage_addFolder('E')
        folderE = folderD['E']
        folderE.manage_addFolder('F')        
        folderF = folderE['F']
        folderF.manage_addDTMLMethod('aTransformer')
        xform = folderF['aTransformer']

        names = self.reg.transformerNames()
        self.assertEquals(len(names),1)
        self.assertEquals(names[0],'A/B/C/D/E/F/aTransformer')

    def test_08(self):
        "Test processor selection machinery"

        names = self.reg.processorNames()
        self.failIf(len(names) == 0)

        selected = self.reg.selected_processor
        self.failUnless(selected in names)

        proc = self.reg.processor()
        self.assertEquals(selected, proc.name)

    ################################################################
    # Test Runner Setup
    ################################################################

if __name__ == '__main__':
    framework(descriptions=1, verbosity=2)
else:
    # While framework.py provides its own test_suite() 
    # method the testrunner utility does not.
    import unittest
    def test_suite():
        suite = unittest.TestSuite()
        suite.addTest(unittest.makeSuite(TransformerRegistryTestCase))
        return suite
